
"""
Test upgrading L{_SubSchedulerParentHook} from version 3 to 4.
"""

# It's the same as upgrading from 2 to 3.
from axiom.test.historic.test_parentHook2to3 import SubSchedulerParentHookUpgradeTests


class SubSchedulerParentHookUpgradeTests(SubSchedulerParentHookUpgradeTests):
    pass
